#include <windows.h>
#include <d3dx9.h>
#include <stdio.h>
#include "../include/StdFps.h"

//Function Declarations
HRESULT InitD3D(HWND hWnd); //Initialize Direct3D Devices
HRESULT InitGeometry();		//Initialize our actual 3D data
void ReleaseD3D();			//Release Direct3D Devices
void ReleaseGeometry();		//Release 3D data
void SetupMatrices();		//Setup projection and world matrices
void Render();				//Render the current frame
LRESULT WINAPI MsgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam); //Windows message handler

//Global DirectX and Windows variables
LPDIRECT3D9		g_pD3D	= NULL; // Used to create the D3DDevice
LPDIRECT3DDEVICE9	g_pd3dDevice	= NULL; // Our rendering device
HWND hWnd;

//Vertex and face information
WORD NumVerts, NumFaces;
LPDIRECT3DVERTEXBUFFER9 g_pVB	= NULL; // Buffer to hold vertices
LPDIRECT3DINDEXBUFFER9	g_pIB	= NULL; // Buffer to hold vertices
struct CUSTOMVERTEX
{
	float x, y, z;
//	DWORD color;
};
#define D3DFVF_CUSTOMVERTEX (D3DFVF_XYZ/*|D3DFVF_DIFFUSE*/)

//Toggle options
bool WireFrame=true, Polys=true, SoftwareMode=true;

//End of header

int WINAPI WinMain(HINSTANCE hInst, HINSTANCE, LPSTR, INT)
{
	//Determine required window size
	RECT ScreenRect={0,0,640,480}; //Window's client dimensions;
	AdjustWindowRect(&ScreenRect, WS_OVERLAPPEDWINDOW, false);

	//Create the window
	WNDCLASSEX wc = {sizeof(WNDCLASSEX), CS_CLASSDC, MsgProc, NULL, NULL, hInst, LoadIcon(hInst, "MAIN_ICON"), NULL, NULL, NULL, "3D Engine", NULL};
	RegisterClassEx(&wc);
	hWnd = CreateWindow("3D Engine", "3D Engine", WS_OVERLAPPEDWINDOW, 100, 100, ScreenRect.right-ScreenRect.left, ScreenRect.bottom-ScreenRect.top, NULL, NULL, hInst, NULL);

	if(SUCCEEDED(InitD3D(hWnd))) //Initialize Direct3D Devices
	{
		if(SUCCEEDED(InitGeometry())) //Initialize our actual 3D data
		{
			//Show the window
			ShowWindow(hWnd, SW_SHOWDEFAULT);
			UpdateWindow(hWnd);

			//Main message loop - Keep going until window is closed
			MSG msg;
			ZeroMemory(&msg, sizeof(msg));
			while(msg.message!=WM_QUIT)
			{
				if(PeekMessage(&msg, NULL, 0U, 0U, PM_REMOVE)) //Process all windows messages before rendering more frames
				{
					TranslateMessage(&msg);
					DispatchMessage(&msg);
				}
				else //If there is no windows message, render another frame
					Render();
			}
			ReleaseGeometry(); //Release 3D data
		}
		ReleaseD3D(); //Release Direct3D Devices
	}

	//Release window data
	UnregisterClass("3D Engine", wc.hInstance);
	return 0;
}

HRESULT InitD3D(HWND hWnd) //Initialize Direct3D Devices
{
	//Initialize DirectX
	if((g_pD3D=Direct3DCreate9(D3D_SDK_VERSION))==NULL)
		return E_FAIL;

	//Initialize DirectX Device
	D3DPRESENT_PARAMETERS d3dpp;
	ZeroMemory(&d3dpp, sizeof(d3dpp));
	d3dpp.Windowed = TRUE;
	d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
	d3dpp.BackBufferFormat = D3DFMT_UNKNOWN;
	d3dpp.EnableAutoDepthStencil = TRUE;
	d3dpp.AutoDepthStencilFormat = D3DFMT_D16;
	d3dpp.PresentationInterval=D3DPRESENT_INTERVAL_IMMEDIATE;
	if(FAILED(g_pD3D->CreateDevice(D3DADAPTER_DEFAULT, SoftwareMode ? D3DDEVTYPE_REF : D3DDEVTYPE_HAL, hWnd, SoftwareMode ? D3DCREATE_SOFTWARE_VERTEXPROCESSING : D3DCREATE_HARDWARE_VERTEXPROCESSING, &d3dpp, &g_pd3dDevice))) //Do everything through software or hardware depending on settings
	{
		g_pD3D->Release();
		return E_FAIL;
	}

	//Set render states
	g_pd3dDevice->SetRenderState(D3DRS_SHADEMODE, D3DSHADE_FLAT); //No shading
	return S_OK;
}

HRESULT InitGeometry() //Initialize our actual 3D data
{
	SetupMatrices(); //Setup projection and world matrices

	//Open the file
	FILE* FileNum=fopen("Models/plane.b3d", "rb");
	if(!FileNum && !(FileNum=fopen("../Models/plane.b3d", "rb")))
	{
		MessageBox(NULL, "File not found", "File Load Error", MB_OK);
		goto Fail1;
	}

	//Determine file length
	fseek(FileNum, 0, SEEK_END);
	DWORD FileLength=ftell(FileNum);
	fseek(FileNum, 0, SEEK_SET);

	//Number of verticies
	fread(&NumVerts, sizeof(NumVerts), 1, FileNum);

	//Create vertex buffer
	void* pVoid;
	if(FAILED(g_pd3dDevice->CreateVertexBuffer(NumVerts*sizeof(CUSTOMVERTEX), 0, D3DFVF_CUSTOMVERTEX, D3DPOOL_DEFAULT, &g_pVB, NULL)))
		goto Fail2;
	if(FAILED(g_pVB->Lock(0, NumVerts*sizeof(CUSTOMVERTEX), (void**)&pVoid, 0)))
		goto Fail3;
	fread(pVoid, NumVerts*sizeof(CUSTOMVERTEX), 1, FileNum);
	g_pVB->Unlock();

	//Number of faces
	fread(&NumFaces, sizeof(NumFaces), 1, FileNum);

	//Create face buffer
	if(FAILED(g_pd3dDevice->CreateIndexBuffer(sizeof(WORD)*3*NumFaces, 0, D3DFMT_INDEX16, D3DPOOL_DEFAULT, &g_pIB, NULL)))
		goto Fail3;
	if(FAILED(g_pIB->Lock(0, sizeof(WORD)*3*NumFaces, (void**)&pVoid, 0)))
		goto Fail4;
	fread(pVoid, sizeof(WORD)*3*NumFaces, 1, FileNum);
	g_pIB->Unlock();
	return S_OK;

	//Each different type of fail needs to close/delete/release more data
	Fail4: g_pIB->Release();
	Fail3: g_pVB->Release();
	Fail2: fclose(FileNum);
	Fail1: return E_FAIL;
}

void ReleaseD3D() //Release Direct3D Devices
{
	if(g_pd3dDevice)
		g_pd3dDevice->Release();
	if(g_pD3D)
		g_pD3D->Release();
}

void ReleaseGeometry() //Release 3D data
{
	if(g_pVB)
		g_pVB->Release();
	if(g_pIB)
		g_pIB->Release();
}

void SetupMatrices() //Setup projection and world matrices
{
	//Camera does not move
	D3DXVECTOR3 vEyePt(0.0f, 3.0f,-5.0f);
	D3DXVECTOR3 vLookatPt(0.0f, 0.0f, 0.0f);
	D3DXVECTOR3 vUpVec(0.0f, 1.0f, 0.0f);
	D3DXMATRIXA16 matView;
	D3DXMatrixLookAtLH(&matView, &vEyePt, &vLookatPt, &vUpVec);
	g_pd3dDevice->SetTransform(D3DTS_VIEW, &matView);

	//Default projection
	D3DXMATRIXA16 matProj;
	D3DXMatrixPerspectiveFovLH(&matProj, D3DX_PI/4, 1.0f, .01f, 10.0f);
	g_pd3dDevice->SetTransform(D3DTS_PROJECTION, &matProj);
}

void RotateWorld() //Rotate the world matrix
{
	//World rotates around x & y axis depending on time
	UINT iTime=GetTickCount()%4000;
	D3DXMATRIXA16 matWorld, matWorld2; //Used to combine matrices
	D3DXMatrixRotationY(&matWorld, iTime*(2.0f*D3DX_PI)/4000.0f);
	D3DXMatrixRotationX(&matWorld2, iTime*(2.0f*D3DX_PI)/2000.0f);
	matWorld=matWorld*matWorld2;
	g_pd3dDevice->SetTransform(D3DTS_WORLD, &matWorld);
}

void Render() //Render the current frame
{
	g_pd3dDevice->Clear(0, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(0,0,0xFF), 1.0f, 0); //Clear frame to blue
	if(SUCCEEDED(g_pd3dDevice->BeginScene()))
	{
		RotateWorld(); //Rotate the model

		//Set render states
		g_pd3dDevice->SetStreamSource(0, g_pVB, 0, sizeof(CUSTOMVERTEX));
		g_pd3dDevice->SetIndices(g_pIB);
		g_pd3dDevice->SetFVF(D3DFVF_CUSTOMVERTEX);
		//Draw mesh as polygons
		if(Polys)
		{
			g_pd3dDevice->SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);
			g_pd3dDevice->SetRenderState(D3DRS_LIGHTING, false);
			g_pd3dDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0, NumVerts, 0, NumFaces);
		}
		//Redraw mesh as wireframe
		if(WireFrame)
		{
			g_pd3dDevice->SetRenderState(D3DRS_FILLMODE, D3DFILL_WIREFRAME);
			g_pd3dDevice->SetRenderState(D3DRS_LIGHTING, true);
			g_pd3dDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0, NumVerts, 0, NumFaces);
		}
		g_pd3dDevice->EndScene();
	}
	g_pd3dDevice->Present(NULL, NULL, NULL, NULL);

	//Display the current FPS and toggle options
	static FPS TheFps;
	char WinText[40];
	TheFps.AddFrame();
	sprintf(WinText, "w:%d, p:%d, s:%d, #:%d, fps:%d", WireFrame&1, Polys&1, SoftwareMode&1, NumFaces, (int)TheFps);
	SetWindowText(hWnd, WinText);
}

LRESULT WINAPI MsgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam) //Windows message handler
{
	if(msg==WM_DESTROY) //If closing window...
	{
		PostQuitMessage(0);
		return 0;
	}
	else if(msg==WM_CHAR) //Keypress for options
		switch((TCHAR)wParam)
		{
			case 'w':WireFrame=!WireFrame; break; //Toggle wireframe mode
			case 'p':Polys=!Polys;break; //Toggle poly mode
			case 's': //Toggle software mode
				SoftwareMode=!SoftwareMode;
				ReleaseGeometry();
				ReleaseD3D();
				if(FAILED(InitD3D(hWnd)) || FAILED(InitGeometry())) //Confirm initialization of Direct3D Devices
					DestroyWindow(hWnd);
		}

	//Default message handler
	return DefWindowProc(hWnd, msg, wParam, lParam);
}