//Standard 3D math operations
#define _PI    ((FLOAT)3.141592654f)

struct TwoDCoord
{
	//2D coordinate definition and constructors
	long x,y;
	inline TwoDCoord(int X, int Y) : x(X),y(Y) {}
	inline TwoDCoord() { }
};

typedef struct Vector3
{
	//Vector definition and constructors
	float x,y,z;
	inline Vector3(float X, float Y, float Z) : x(X),y(Y),z(Z) {}
	inline Vector3() { }

	//Simple vector arithmetic
	Vector3 operator - (Vector3& Vec2) { return Vector3(x-Vec2.x, y-Vec2.y, z-Vec2.z); }
	Vector3 operator + (Vector3& Vec2) { return Vector3(x+Vec2.x, y+Vec2.y, z+Vec2.z); }
	Vector3 operator * (Vector3& Vec2) { return Vector3(y * Vec2.z - z * Vec2.y, z * Vec2.x - x * Vec2.z, x * Vec2.y - y * Vec2.x); } //Cross
} Vertex;

struct Vector4 { float x,y,z,w; };
struct Face { unsigned short VertexIndices[3]; };
struct Matrix {
	union
	{
		struct
		{
			float _11, _12, _13, _14;
			float _21, _22, _23, _24;
			float _31, _32, _33, _34;
			float _41, _42, _43, _44;
		};
		float m[4][4];
	};
	inline Matrix(float M[16]) { for(int i=0;i<16;i++) ((float*)m)[i]=M[i];}
	inline Matrix() { }
	Matrix operator * (Matrix& Mat2);
};

Matrix* MatrixMultiply(Matrix *Out, Matrix*, Matrix*);
Vector4 MatrixVectorMultiply(Matrix *A, Vector3 *B);
Vector3 MatrixVectorProject(Matrix*, Vector3*);
float Vec3Dot(Vector3*, Vector3*);
Vector3 NormalizeVec(Vector3);
TwoDCoord GetTwoDCoord(Vertex*, Matrix* ProjectionMatrix, int Width, int Height);
Matrix* MatrixLookAtLH(Matrix *Out, Vector3 *Eye, Vector3 *At, Vector3 *Up);
Matrix* MatrixPerspectiveFovLH(Matrix *Out, float fovy, float Aspect, float zn, float zf);
Matrix* MatrixRotationX(Matrix *Out, float Angle);
Matrix* MatrixRotationY(Matrix *Out, float Angle);
Matrix* MatrixRotationZ(Matrix *Out, float Angle);
