//Calculates FPS using an average over the last 2 seconds
#include <string.h> //memset, strcat
#include <stdlib.h> //gcvt, ltoa
#include <time.h> //clock
#include "StdFps.h"

FPS::FPS() //Constructor
{
	TotalFrames=0;
	ResetFrames();
	LastFpsStringPointer=NULL;
}

FPS::~FPS() //Destructor
{
	if(LastFpsStringPointer!=NULL)
		delete[] LastFpsStringPointer;
	LastFpsStringPointer=NULL;
}

void FPS::AddFrame() //Call when a new frame has been drawn
{
	FrameOn++;
	TotalFrames++;
}

void FPS::ResetFrames() { FrameOn=StoreTicks=OldTicks=LastTotalFrames=0; } //Reset fps system
DWORD FPS::GetTotalFrames() { return TotalFrames; } //Get the total number of frames the program has drawn

float FPS::GetFps() //Returns the current average FPS as a float
{
	DWORD CurrentTicks=clock()*1000/CLOCKS_PER_SEC; //Milliseconds since start of program
	float AverageFPS=(1000.0f*FrameOn/(float)(CurrentTicks-OldTicks));

	if((CurrentTicks-OldTicks)>2000) //After 2 seconds is up, move the number of frames from the last second *only* into a buffer to keep the frame rate accurate
	{
		FrameOn-=LastTotalFrames;
		LastTotalFrames=FrameOn;
		OldTicks=StoreTicks;
		StoreTicks=CurrentTicks;
	}
	return AverageFPS;
}

char* FPS::GetFpsString(int NumberOfDecimals) //Returns the current average FPS as a string with the provided accuracy
{
	if(LastFpsStringPointer!=NULL) //Delete the last fps string
		delete[] LastFpsStringPointer;

	char *PrintFPS = new char[9+NumberOfDecimals];
	memset(PrintFPS,0,9+NumberOfDecimals);
	if(NumberOfDecimals==0)
	{
		DWORD AverageFPS=(DWORD)GetFps();
		ltoa(AverageFPS, PrintFPS, 10);
	}
	else
	{
		float AverageFPS=GetFps();
		gcvt(AverageFPS, NumberOfDecimals, PrintFPS);
	}
	strcat(PrintFPS, " fps");
	LastFpsStringPointer=PrintFPS;
	return PrintFPS;
}